#pragma once

#define MAX_PROFILER_SAMPLES 500

/*! Due to some MinGW-w32api oddness, we must redefine the OS version here.
 *  This is mainly so GetGuiResources will work.
 */
#undef WINVER
#define WINVER 0x0501

#include "StatsWriter.hh"
#include "RingBuffer.hh"

/*! Because windows.h includes a lot of other files by default, the resulting
 *  object code can be quite large even if most of it isn't used. For this
 *  reason these defines are provided to remove the largest unused portions of
 *  the code at compile time.
 */
#define WIN32_LEAN_AND_MEAN //!< Removes most of windows.h.
#define VC_EXTRALEAN //!< Slims down windows.h a lot.
#define NOGDI //!< Removes some unused GDI support from windows.h.
#define NOSOUND //!< Removes media support from windows.h.

#include "windows.h"
#include "stdint.h"

#include <string>
#include <map>
#include <utility> // For pair.

using namespace std;

//! A lossy recorded time sample.
/*! A time sample is stored in the profiler until it can be manipulated by the
 *  \c Profiler and output by a \c StatsWriter. After a \c Sample is merged into
 *  a \c Record it is discarded. The \c Sample takes accurate measurements, but
 *  the \c Record stores percentages and averages.
 */
class Record {
   private:
      pair<HWND, uint32_t> pIdent; //!< Identifier of the sample.
      uint32_t uCalls; //!< Number of calls this frame.
      uint32_t uData; //!< Number of percentage values stored.
      uint64_t dStart; //!< Starting clock time.
      double dTotal; //!< Total time recorded.
      double dAvgPc; //!< Mean average percentage of message loop time taken up.
      double dMinPc; //!< Minimum percentage of message loop time taken up.
      double dMaxPc; //!< Maximum percentage of message loop time taken up.
      double dPercent; //!< Total percentage time.
      bool bValid; //!< Indicates if the sample is valid to use.
      bool bOpen; //!< Indicates if the sample is currently being profiled.
   public:
      Record();
      ~Record();
      const bool is_valid() const;
      const bool is_open() const;
      const bool is_closed() const;
      const pair<HWND, uint32_t> &get_ident() const;
      const uint64_t get_start() const;
      const double get_average() const;
      const uint32_t get_calls() const;
      const double get_min() const;
      const double get_max() const;
      const double get_percent() const;
      const double get_total() const;
      const uint32_t get_data() const;
      void open();
      void close();
      void start();
      void make_valid();
      void make_invalid();
      void set_ident(const pair<HWND, uint32_t>);
      void set_calls(const uint32_t);
      void set_total(const double);
      void set_percent(const double);
      void set_average(const double);
      void set_min(const double);
      void set_max(const double);
      void add_data(const uint32_t = 1);
      void add_call(const uint32_t = 1);
      void add_total(const double dTime = 1);
      void reset();
};

//! The heart of the Profiler system.
/*! The design of this system component is inspired by the \c Profiler class in
 *  the "Enginuity" articles by Richard Fine. The Profiler works by passively
 *  checking the RDTSC rather than timing how long a loop is taking. As well as
 *  the Profiler's measure of the entire profiling run, Samples are constructed
 *  and destructed at strategic points in the host program's code.
 */
class Profiler {
   private:
      static const string get_local_time();
      static const uint64_t filetime_to_uint64_t(const FILETIME);
      uint64_t dStart; //<! Start tick for the whole profiling run.
      uint64_t dEnd; //<! End tick for the whole profiling run.
      uint32_t uGdiObjects; //<! The number of GDI objects at the start.
      uint32_t uUserObjects; //<! The number of user objects at the start.
      bool bRunning; //<! Is the profiler running?
      static uint64_t uTicksPerSecond; //<! Ratio of ticks to time.
      uint64_t uLastCpuKernel; //<! Last CPU kernel mode measurement.
      uint64_t uLastCpuUser; //<! Last CPU user mode measurement.
      uint64_t uLastKernel; //<! Last process kernel mode measurement.
      uint64_t uLastUser; //<! Last process user mode measurement.
   public:
      Profiler();
      ~Profiler();
      void output(const StatsWriter *);
      static const uint64_t get_time();
      const double get_total_run_time() const;
      const bool is_running() const;
      void calculate_stats();
      void record_resource_usage();
      const double get_current_cpu_usage();
      const uint32_t get_current_ram_usage();
      const uint32_t get_current_vmm_usage();
      const uint16_t get_current_gdi_objects();
      const uint16_t get_current_user_objects();
      const uint16_t get_stored_gdi_objects();
      const uint16_t get_stored_user_objects();
      static const uint64_t get_ticks_per_second();
      static inline volatile uint64_t get_cycle_count();
      static const uint32_t get_cpu_speed();
      BlackHoleRingBuffer<double, 100> rCpu; //!< 100 CPU percentage readings.
      BlackHoleRingBuffer<uint32_t, 100> rRam; //!< 100 RAM usage readings.
      BlackHoleRingBuffer<uint32_t, 100> rVmm; //!< 100 virtual memory readings.
      BlackHoleRingBuffer<uint16_t, 100> rGdi; //!< 100 GDI object readings.
      BlackHoleRingBuffer<uint16_t, 100> rUser; //!< 100 user object readings.
};

//! A time sample.
/*! This transitory state between profiled code and \c Record performs some
 *  vital functions. For instance, the \c Sample enables code to be timed.
 */
class Sample {
   private:
      static uint64_t dBegin; //<! Starting tick for this sample.
      static uint64_t dEnd; //<! Ending tick for this sample.
      static int32_t iLastSample; //<! The index of the last sample.
      static int32_t iOpenSamples; //<! The number of open samples.
      int32_t iSampleIndex; //<! The index of this sample.
      int32_t iParentIndex; //<! Unused remnant of the original design.
   public:
      void init(HWND, uint32_t);
      void shutdown();
      static void reset_all();
      static void reset(const pair<HWND, uint32_t>);
      void start(double);
      static uint32_t get_total_samples();
      static Profiler *pProfiler; //<! A global pointer to our \c Profiler.
      static Record rSamples[MAX_PROFILER_SAMPLES]; //<! The \c Record store.
};
