#pragma once

#include "stdint.h"

#include <cassert>

/*! \brief A simple ring buffer interface.
 *
 * The idea is to create a fast buffer of virtually infinite size, provided the
 * read and write pointers never meet.
 */
class RingBase {
   public:
      // Ugly hack to get around the need for virtual template functions. :(
      // To fix later.
      //! Write to the buffer.
      virtual RingBase &operator<<(const uint32_t) { return *this; }
      //! Read from the buffer.
      virtual RingBase &operator>>(uint32_t &) { return *this; }
      //! Write to the buffer.
      virtual RingBase &operator<<(const uint16_t) { return *this; }
      //! Read from the buffer.
      virtual RingBase &operator>>(uint16_t &) { return *this; }
      //! Write to the buffer.
      virtual RingBase &operator<<(const uint8_t) { return *this; }
      //! Read from the buffer.
      virtual RingBase &operator>>(uint8_t &) { return *this; }
      //! Take some action if the buffer is full.
      virtual void check_full_write() = 0;
      //! Take some action if the buffer is empty.
      virtual void check_empty_read() = 0;
      //! Return the number of readable data.
      virtual const uint32_t size() const = 0;
      //! Is the buffer empty?
      virtual const bool is_empty() const = 0;
      //! Is the buffer full?
      virtual const bool is_full() const = 0;
};

/*! \brief The traditional 'infinite' ring buffer.
 *
 *  This is an implementation of the traditional "ring buffer" model. The buffer
 *  has a fixed size and its bounds SHALL NOT be violated, by which I mean
 *  assertions are manifest if the buffer is read while empty or written while
 *  full. The pragma directive ensures the carefully ordered data members are
 *  aligned to 1-byte boundaries, and so are accessed much more quickly than
 *  data members arranged in memory any old how.
 * \note The terms 'full' and 'empty' are deemed adequate to describe the
 *       situations in which the write buffer has caught up to the read buffer
 *       and vice versa, respectively.
 */
#pragma pack(push, 1)
template<class T, uint32_t uSize>
class RingBuffer : public RingBase {
   protected:
      /*! \brief Increment one of the pointers, taking account of wrapping.
       * \param[in,out] uPtr The pointer to increment.
       */
      void inc_wrap(uint32_t &uPtr) {
         if (uPtr == uSize - 1) {
            uPtr = 0;
         } else {
            ++uPtr;
         }
      }
      T tBuffer[uSize]; //!< \brief The data buffer itself.
      uint32_t uReadPtr; //!< \brief The read pointer.
      uint32_t uWritePtr; //!< \brief The write pointer.
      bool bFull; //!< \brief The 'full' indicator.
      bool bEmpty; //!< \brief The 'empty' indicator.
   public:
      //! A default constructor which fills the data buffer with zeros.
      RingBuffer() : uReadPtr(0), uWritePtr(0), bFull(false), bEmpty(true) {
         zerofill();
      }
      //! Ensure a full buffer is never written.
      void check_full_write() {
         assert(bFull == false && "Write to a full buffer!");
      }
      //! Ensure an empty buffer is never read.
      void check_empty_read() {
         assert(bEmpty == false && "Read from an empty buffer!");
      }
      //! Write an item into the buffer.
      RingBuffer &operator<<(const T tAdd) {
         check_full_write();
         tBuffer[uWritePtr] = tAdd;
         inc_wrap(uWritePtr);
         bEmpty = false;
         if (uReadPtr == uWritePtr) { // Buffer is now full.
            bFull = true;
         }
         return *this;
      }
      //! Read an item from the buffer.
      RingBuffer &operator>>(T &tRetVal) {
         check_empty_read();
         tRetVal = tBuffer[uReadPtr];
         inc_wrap(uReadPtr);
         bFull = false;
         if (uReadPtr == uWritePtr) { // Buffer is now empty.
            bEmpty = true;
         }
         return *this;
      }
      //! Flood the buffer with zeros.
      void zerofill() {
         flood(0);
      }
      //! Write the given value to all indices.
      void flood(const T &tFlood) {
         for (uWritePtr = uSize; uWritePtr--; ) {
            tBuffer[uWritePtr] = tFlood;
         }
         uReadPtr = 0;
         uWritePtr = 0;
      }
      //! Return the amount of readable data.
      const uint32_t size() const {
         uint32_t uRetVal = 0;
         if (uReadPtr != uWritePtr) {
            if (uReadPtr < uWritePtr) {
               uRetVal = uWritePtr - uReadPtr;
            } else {
               uRetVal = uSize - uReadPtr + uWritePtr;
            }
         } else {
            if (bFull == true) {
               uRetVal = uSize;
            }
         }
         return uRetVal;
      }
      //! Is the buffer empty?
      const bool is_empty() const {
         return bEmpty;
      }
      //! Is the buffer full?
      const bool is_full() const {
         return bFull;
      }
};

/*! \brief The 'black hole' model ring buffer.
 *
 *  This implementation is the same as RingBuffer except that the rules
 *  regarding fullness and emptiness are relaxed. An empty buffer may be read,
 *  and a full buffer may be written. These conditions might be desirable when,
 *  for instance, keeping the last 100 records of some variable quantity. When
 *  the write pointer catches the read pointer, it moves the read pointer
 *  forward to match its own movement. Thus the read pointer will always point
 *  at the oldest datum in the ring. Since it keeps on eating data regardless, I
 *  called it a "black hole" ring buffer. The analogue eats matter which powers
 *  its periodic gamma bursts. It seemed appropriate somehow.
 * \note I would have this class as a derivative of RingBuffer, but the template
 *  parameters somehow prevent inheritance.
 */
template<class T, uint32_t uSize>
class BlackHoleRingBuffer : public RingBase {
   protected:
      /*! \brief Increment one of the pointers, taking account of wrapping.
       * \param[in,out] uPtr The pointer to increment.
       */
      void inc_wrap(uint32_t &uPtr) {
         if (uPtr == uSize - 1) {
            uPtr = 0;
         } else {
            ++uPtr;
         }
      }
      T tBuffer[uSize]; //!< \brief The data buffer itself.
      uint32_t uReadPtr; //!< \brief The read pointer.
      uint32_t uWritePtr; //!< \brief The write pointer.
      bool bFull; //!< \brief The 'full' indicator.
      bool bEmpty; //!< \brief The 'empty' indicator.
   public:
      //! A default constructor which fills the data buffer with zeros.
      BlackHoleRingBuffer() : uReadPtr(0), uWritePtr(0), bFull(false),
                              bEmpty(true) {
         zerofill();
      }
      //! If a full buffer is about to be written, move the read pointer.
      void check_full_write() {
         if (bFull == true) {
            inc_wrap(uReadPtr);
         }
      }
      //! Do nothing. We don't care if an empty buffer is read.
      void check_empty_read() {
         // Do nothing!
      }
      //! Write an item into the buffer.
      BlackHoleRingBuffer &operator<<(const T tAdd) {
         check_full_write();
         tBuffer[uWritePtr] = tAdd;
         inc_wrap(uWritePtr);
         bEmpty = false;
         if (uReadPtr == uWritePtr) { // Buffer is now full.
            bFull = true;
         }
         return *this;
      }
      //! Read an item from the buffer.
      BlackHoleRingBuffer &operator>>(T &tRetVal) {
         check_empty_read();
         tRetVal = tBuffer[uReadPtr];
         inc_wrap(uReadPtr);
         bFull = false;
         if (uReadPtr == uWritePtr) { // Buffer is now empty.
            bEmpty = true;
         }
         return *this;
      }
      //! Flood the buffer with zeros.
      void zerofill() {
         flood(0);
      }
      //! Write the given value to all indices.
      void flood(const T &tFlood) {
         // Write the value to all indices. A reset in other words.
         for (uWritePtr = uSize; uWritePtr--; ) {
            tBuffer[uWritePtr] = tFlood;
         }
         uReadPtr = 0;
         uWritePtr = 0;
      }
      //! Return the amount of readable data.
      const uint32_t size() const {
         uint32_t uRetVal = 0;
         if (uReadPtr != uWritePtr) {
            if (uReadPtr < uWritePtr) {
               uRetVal = uWritePtr - uReadPtr;
            } else {
               uRetVal = uSize - uReadPtr + uWritePtr;
            }
         } else {
            if (bFull == true) {
               uRetVal = uSize;
            }
         }
         return uRetVal;
      }
      //! Is the buffer empty?
      const bool is_empty() const {
         return bEmpty;
      }
      //! Is the buffer full?
      const bool is_full() const {
         return bFull;
      }
};
#pragma pack(pop)
