#include "../include/Hooker.hh"

#include "../include/StatsWriter.hh"

// Window enumerators //////////////////////////////////////////////////////////

//! Enumerate windows until a window owned by this process is encountered.
/*! The \c EnumerateChildWindows() API is used to call each descendent of the
 *  specified window with a <tt>BOOL (*)(HWND, LPARAM)</tt> function.
 *  This function stores the window handle and stops enumeration if the window
 *  belongs to the current process.
 *  \param[in] hWnd A window handle.
 *  \return A Microsoft-defined boolean value. Unusually, false indicates
 *          success.
 */
BOOL EnumFirst(HWND hWnd, LPARAM) {
   // If the window is owned by this process, store it and stop enumerating.
   bool bRetVal = true; // True here means 'continue'. False means 'stop'.
   DWORD dHwndProcess = 0;
   GetWindowThreadProcessId(hWnd, &dHwndProcess);
   if (uProcess == dHwndProcess) {
      hFirst = hWnd;
      bRetVal = false;
   }
   return bRetVal;
}

//! Enumerate windows and hook any new ones belonging to this process.
/*! This function is very similar to \c EnumFirst() except that the enumeration
 *  stops when there aren't any more windows. If a window belonging to this
 *  process is found that isn't in the map, then hook it and add it to the map.
 *  \param[in] hWnd A window handle.
 *  \return A Microsoft-defined boolean value.
 *  \note Checking if the window is in the list has a time complexity of O(n),
 *        whereas checking if this process owns the window has a time complexity
 *        of O(1), so to save time the window's ownership is checked first.
 *  \bug This functionality is not yet implemented. See \c HookProc().
 */
BOOL EnumChildren(HWND hWnd, LPARAM) {
   // If the window is owned by this process, check if it's already subclassed.
   // If not, subclass it and add to the list.
   DWORD dHwndProcess = 0;
   GetWindowThreadProcessId(hWnd, &dHwndProcess);
   if (uProcess == dHwndProcess) {
      if (mWindows.find(hWnd) == mWindows.end()) {
         mWindows[hWnd] = SetWindowLong(hWnd, GWL_WNDPROC, (long)HookProc);
      }
   }
   return true;
}

// Entry Point helpers /////////////////////////////////////////////////////////

//! Hook the first window.
/*! \return A boolean true value. It's a slightly odd thing to do in a DWORD,
 *          but it's not \e wrong, because a bool type resolves to an integer
 *          anyway.
 *  \bug \c EnumChildWindows() requires an explicit cast to WNDENUMPROC, even if
 *       the function is of that type anyway. If the cast is inadvertently left
 *       out, the compiler gives an error of the form "invalid conversion from
 *       'X' to 'X'". While highly amusing, it is presumably a MinGW bug, and
 *       therefore beyond the scope of this project.
 */
DWORD WINAPI Hook(void *) {
   uProcess = GetCurrentProcessId();
   // At first we just need to subclass one window - any window - of the current
   // process, to bootstrap our window subclassing algorithm. The rest we can
   // subclass later. For now, we'll just grab the first one we come across.
   while (hFirst == 0) {
      EnumChildWindows(hDesktop, (WNDENUMPROC)EnumFirst, 0);
   }
   mWindows[hFirst] = SetWindowLong(hFirst, GWL_WNDPROC, (long)HookProc);
   return true; // Got to return something. :/
}

//! Unhook all windows.
/*! If this isn't done and the target process continues with its windows intact,
 *  the process is 100% guaranteed to crash almost immediately. The message pump
 *  would be calling a function that isn't there any more.
 *  \return A boolean true value. It's a slightly odd thing to do in a DWORD,
 *          but it's not \e wrong, because a bool resolves to an integer anyway.
 */
DWORD WINAPI Unhook(void *) {
   window_map::iterator wIter = mWindows.begin();
   for ( ; wIter != mWindows.end(); wIter++) {
      if ((*wIter).first != 0) {
         SetWindowLong((*wIter).first, GWL_WNDPROC, (*wIter).second);
      }
   }
   return true; // Got to return something. :/
}

// Entry Point /////////////////////////////////////////////////////////////////

//! The default entry point.
/*! The entry point of a DLL is quite a fragile place to be, due to the lock
 *  Windows obtains on the DLL loading mechanism. Therefore nothing is done here
 *  that might cause a deadlock, such as loading another library or causing
 *  another library to be loaded. Since \c DisableThreadLibraryCalls() is called
 *  early on, the function doesn't have to process the DLL_THREAD_ATTACH and
 *  DLL_THREAD_DETACH reason codes at all, so they can be ignored. \c DllMain()
 *  is called implicitly by \c LoadLibrary() and \c FreeLibrary(), never
 *  explicitly by a program.
 *  \param[in] hInst A process instance.
 *  \param[in] uReason A \c DllMain() reason code.
 *  \param[in] vPtr Despite being described as "reserved", it's usable already.
 *  \return Returns a boolean value which indicates success or failure.
 */
extern "C"
bool APIENTRY DllMain(HINSTANCE hInst, uint32_t uReason, void *vPtr) {
   bool bRetVal = TRUE;
   HANDLE hThread = 0;
   if (uReason == DLL_PROCESS_ATTACH) {
      DisableThreadLibraryCalls(hInst);
      // It's ok to call CreateThread from DllMain because it lives in
      // kernel32.dll. Kernel32.dll functions are safe, according to MS.
      hThread = CreateThread(0, 0, &Hook, 0, 0, 0);
      // Close the handle immediately it's created. Remember this only deletes
      // the handle, not the thread. The thread continues regardless.
      CloseHandle(hThread);
      // Start the profiler.
      Sample::pProfiler = new Profiler();
   }
   if (uReason == DLL_PROCESS_DETACH) {
      // Stop the profiler, calculate results and output a file.
      Sample::pProfiler->calculate_stats();
      StatsWriter *sWriter = new XmlStatsWriter();
      Sample::pProfiler->output(sWriter);
      delete sWriter;
      delete Sample::pProfiler;
      Sample::pProfiler = 0;
      // DllMain is called with the reason code DLL_PROCESS_DETACH for two
      // reasons: a) because the process it's attached to is exiting, or
      // b) FreeLibrary was called. The content of vPtr will be 0 if
      // FreeLibrary was called; not otherwise.
      if (vPtr == 0) {
         CloseHandle(CreateThread(0, 0, &Unhook, 0, 0, 0));
      }
      // Otherwise just drop everything, because the target process is going
      // away anyway. By this point all windows have been destroyed, so there is
      // no danger of a crash. Well... not from that anyway.
   }
   return bRetVal;
}

// Hook procedure //////////////////////////////////////////////////////////////

//! The hook procedure.
/*! The hook procedure's address replaces the window procedure's address in all
 *  hooked windows. The window procedure's address is stored in the \c mWindows
 *  map, to be called at the end of the hook procedure.
 *  \param[in] hWnd A window handle.
 *  \param[in] uMsg A message identifier.
 *  \param[in] wParam The smaller of two data parameters.
 *  \param[in] lParam The larger of two data parameters.
 *  \return Windows decrees that a rather vaguely defined four-byte integer 
 *          value be returned. An \c LRESULT can be cast to almost any type.
 *  \todo The \c EnumThreadWindows() function should be called from here.
 *  \bug When enabled, the \c EnumThreadWindows() function call crashes.
 *  \note The hook procedure is always executed \e before the corresponding
 *        window procedure. This means we can 'straddle' the window procedure
 *        and time it.
 */
LRESULT CALLBACK HookProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
   WNDPROC wProc = reinterpret_cast<WNDPROC>(mWindows[hWnd]);
   LRESULT lRetVal = 0;
//   if (++uMessages % MESSAGE_EPOCH == 0) {
      // Enumerate this process's windows. If there's a new one, subclass it
      // and insert in the internal window list. Try and reduce the time
      // complexity of this process as much as possible, because obviously we
      // need to be unobtrusive.
      // After a test, I found that EnumChildWindows deals with ALL the same
      // windows as EnumWindows, as long as the first parameter is the
      // desktop window. Therefore we don't need EnumWindows.
      // TO DO: EnumThreadWindows?
      // Best case: O(n). Worst case: O(n^2).
      // Every MESSAGE_EPOCH messages, enumerate windows.
// TO DO: Fix crash bug in this function.
//      EnumChildWindows(hDesktop, (WNDENUMPROC)EnumChildren, 0);
//   }
   Sample *sCurrent = new Sample();
   sCurrent->init(hWnd, uMsg);
   lRetVal = CallWindowProc(wProc, hWnd, uMsg, wParam, lParam);
   sCurrent->shutdown();
   delete sCurrent;
   return lRetVal;
}
