/*! \mainpage ProgMon.exe
 *  \brief Generates statistics about other programs.
 *  \author Adam J Richardson
 *  \version 0.1
 *  \date 2006-2007
 *
 *  \section intro_sec Introduction
 *  This application monitors other programs' window messages and records useful
 *  statistics. It is meant for comparing programs. For instance, the amount of
 *  time a program spends handling WM_PAINT could be significant in determining
 *  how efficient it is.
 *
 *  ProgMon.exe tries to be as passive as possible in order to not influence the
 *  results.
 */

#include "../include/Main.hh"

//! Converts a string to lower case.
/*! \param[in] sToLower The mixed case string.
 *  \return The lower case string.
 */
string str_to_lower(string sToLower) {
   string sCaps = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
   string::iterator sIter = sToLower.begin();
   for ( ; sIter < sToLower.end(); sIter++) {
      if (sCaps.find(*sIter) != string::npos) {
         *sIter = tolower(*sIter);
      }
   }
   return sToLower;
}

//! Return an open process handle.
/*! The ToolHelp32 API allows any Windows version since NT 4 to enumerate its
 *  own processes. When a process with an executable named \a sExeName is found,
 *  the function creates a handle to that process and returns it.
 *  \param[in] sExeName The name of the executable to seek.
 *  \return An open handle to a process.
 *  \note The found process is not necessarily the one sought. There may be two
 *  processes with the same executable name, or even loaded from the same
 *  executable.
 */
HANDLE get_process_handle(string sExeName) {
   HANDLE hRetVal = 0;
   try {
      PROCESSENTRY32 pEntry = {sizeof(PROCESSENTRY32), 0, 0, 0, 0, 0, 0, 0, 0,
                               ""};
      HANDLE hSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
      if (hSnap == INVALID_HANDLE_VALUE) {
         throw string("CreateToolhelp32Snapshot: couldn't take snapshot.");
      }
      string sTemp = "";
      size_t sPos = 0;
      if (Process32First(hSnap, &pEntry) == TRUE) {
         // Loop time complexity: O(n)
         do {
            sTemp = pEntry.szExeFile; // Compare only the filename portion.
            if ((sPos = sTemp.find('\\')) != string::npos) {
               sTemp = sTemp.substr(sPos + 1);
            }
            cout << "Found " << sTemp << ".";
            if (sExeName == str_to_lower(sTemp)) {
               cout << " That's the one!" << endl;
               hRetVal = OpenProcess(PROCESS_ALL_ACCESS, false,
                                     pEntry.th32ProcessID);
               if (hRetVal == 0) {
                  throw string("OpenProcess: access denied.");
               }
               break;
            }
            cout << endl;
         } while (Process32Next(hSnap, &pEntry));
      } else {
         throw string(GetLastError() == ERROR_NO_MORE_FILES ? "Process32First:"
                      " No processes exist!" : "Process32First: The snapshot "
                      "contains no process information!");
      }
   } catch (...) {
      throw;
   }
   return hRetVal;
}

//! Inject the hook library.
/*! Since the prototype of LoadLibrary is the same as LPTHREAD_START_ROUTINE, a
 *  DLL can be attached to a process by means of the CreateRemoteThread API.
 *  First an open handle to the process must be obtained and passed into the
 *  function in \a hProcess. A few bytes of memory are assigned to the new
 *  thread for the contents of \a sDll, and then the remote thread is created.
 *  If the thread terminated with no error code, the DLL is now hooked to the
 *  remote process. Finally, the process runs DllMain with the
 *  'DLL_PROCESS_ATTACH' reason code.
 *  \pre The handle \e must be open.
 *  \param[in] hProcess An open handle to a process.
 *  \param[in] sDll The DLL to load. A relative path should work fine.
 */
void inject(HANDLE hProcess, string sDll) {
   try {
      cout << "Injecting hook into handle " << hProcess << "..." << endl;
      uint32_t uError = 0;
      size_t sLen = sDll.length();
      void *vPtr = 0;
      HMODULE hKernel = GetModuleHandle("Kernel32");
      HANDLE hThread = 0;
      if (hKernel != 0 && hProcess != 0) {
         vPtr = VirtualAllocEx(hProcess, 0, sLen, MEM_COMMIT, PAGE_READWRITE);
         if (vPtr == 0) {
            throw string("VirtualAllocEx: couldn't allocate process memory.");
         }
         uint32_t uWait = 0;
         DWORD dResult = 0;
         uError = WriteProcessMemory(hProcess, vPtr, sDll.c_str(), sLen, 0);
         if (uError == 0) {
            throw string("WriteProcessMemory: couldn't access allocated "
                         "memory.");
         }
         LPSTART tStart = (LPSTART)GetProcAddress(hKernel, "LoadLibraryA");
         if (tStart == 0) {
            throw string("GetProcAddress: couldn't find LoadLibrary function.");
         }
         hThread = CreateRemoteThread(hProcess, 0, 0, tStart, vPtr, 0, 0);
         if (hThread == 0) {
            throw string("CreateRemoteThread: couldn't create remote thread.");
         }
         if (hThread != 0) {
            // Execute LoadLibrary("Hooker.dll") in the remote process.
            uWait = WaitForSingleObject(hThread, 10000); // Time out in 10 secs.
            if (uWait == WAIT_FAILED) {
               throw string("WaitForSingleObject: couldn't wait on thread.");
            }
            if (uWait == WAIT_ABANDONED) {
               throw string("WaitForSingleObject: mutex object locked.");
            }
            // If the load succeeded, dResult is a HINSTANCE to the DLL. Joy.
            uError = GetExitCodeThread(hThread, &dResult);
            if (uError == 0) {
               throw string("GetExitCodeThread: couldn't get return value.");
            }
            uError = CloseHandle(hThread);
            if (uError == 0) {
               throw string("CloseHandle: couldn't release handle.");
            }
         }
         uError = VirtualFreeEx(hProcess, vPtr, 0, MEM_RELEASE);
         if (uError == 0) {
            throw string("VirtualFreeEx: couldn't release process memory.");
         }
         // LoadLibrary blocked. A ten second delay counts as blocking.
         if (uWait == WAIT_TIMEOUT) {
            char cTemp[64];
            sprintf(cTemp, "inject: hooker blocked. Reason: Error %ld.",
                    GetLastError());
            throw string(cTemp);
         }
         // LoadLibrary failed if the return value is 0.
         if (dResult == 0) {
            char cTemp[64];
            sprintf(cTemp, "inject: hooker failed. Reason: Error %ld.",
                    GetLastError());
            throw string(cTemp);
         }
      }
   } catch (...) {
      throw;
   }
}

//! Monitor the process.
/*! Using \c get_process_handle(), the function locates an appropriate process.
 *  When found, an open handle to the process can then be passed to the
 *  \c inject() function which will attempt to attach the hook library to the
 *  process.
 *  \param[in] sCommand The command to pass to system.
 *  \return An 8-bit signed integer. Based on whether the hook succeeded, return
 *          a value to return to whatever executed ProgMon.exe.
 */
const int8_t monitor(const string sCommand) {
   string sShort = sCommand;
   size_t sPos = sCommand.rfind('\\');
   if (sPos != string::npos) {
      sShort = sShort.substr(++sPos);
   }
   sShort = str_to_lower(sShort);
   HANDLE hOpen = 0;
   int8_t iRetVal = -1;
   try {
      // Loop time complexity: O(n^2)
      // Find the process handle.
      while ((hOpen = get_process_handle(sShort)) == 0) {
         Sleep(1); // Sleeping prevents CPU lock if there is no such process.
      }
      char cPath[256];
      GetModuleFileName(0, (char *)&cPath, 256);
      string sPath = cPath;
      size_t sPos = sPath.rfind('\\');
      if (sPos != string::npos) {
         sPath = sPath.substr(0, sPos + 1);
      }
      sPath += "Hooker.dll";
      // Attempt to inject our "hooker" DLL into the process.
      inject(hOpen, sPath);
      CloseHandle(hOpen);
      iRetVal = 0;
   } catch (...) {
      throw;
   }
   return iRetVal;
}

//! Try to monitor a process.
/*! Attempt to monitor the process with the given name, and indicate success or
 *  failure by a pop-up message.
 *  \param[in] iArgs The number of command line arguments passed.
 *  \param[in] cArgs An array of C-style strings to be parsed as arguments.
 *  \return A standard integer value indicating success or failure.
 */
int main(int iArgs, char **cArgs) {
   uint8_t uRetVal = 0;
   if (iArgs > 1) {
      try {
         // Attempt to monitor the process identified by cArgs[1].
         monitor(cArgs[1]);
         // This messagebox prevents misuse of the monitor if somehow the binary
         // is leaked.
         MessageBox(0, "Monitoring started!", "ProgMon", 0);
      } catch (string &sRetVal) {
         cout << sRetVal << endl;
         uint32_t uError = GetLastError();
         if (uError != 0) {
            cout << "Reason: system error #" << uError << "." << endl;
         } else {
            cout << "There is no error code." << endl;
         }
         uRetVal = 1;
      }
   }
   return uRetVal;
}
