#include "../include/Profiler.hh"

#include <fstream>
#include "Psapi.h"

// TO DO: doxygen tags in this file. Also StatsWriter.cc remains.

// Record class ////////////////////////////////////////////////////////////////

Record::Record() : uData(0), dAvgPc(-1), dMinPc(-1), dMaxPc(-1), bValid(false) {
}

Record::~Record() {
}

const bool Record::is_valid() const {
   return bValid;
}

const bool Record::is_open() const {
   return (bOpen == true);
}

const bool Record::is_closed() const {
   return (bOpen == false);
}

void Record::open() {
   bOpen = true;
}

void Record::close() {
   bOpen = false;
}

const double Record::get_total() const {
   return dTotal;
}

const uint32_t Record::get_data() const {
   return uData;
}

const pair<HWND, uint32_t> &Record::get_ident() const {
   return pIdent;
}

const uint64_t Record::get_start() const {
   return dStart;
}

const double Record::get_average() const {
   return dAvgPc;
}

const uint32_t Record::get_calls() const {
   return uCalls;
}

const double Record::get_min() const {
   return dMinPc;
}

const double Record::get_max() const {
   return dMaxPc;
}

const double Record::get_percent() const {
   return (100.0f * dAvgPc * uData) / (Profiler::get_ticks_per_second() *
                                       Sample::pProfiler->get_total_run_time());
}

void Record::start() {
   dStart = Profiler::get_time();
}

void Record::make_valid() {
   bValid = true;
}

void Record::make_invalid() {
   bValid = false;
}

void Record::add_call(const uint32_t uNum) {
   uCalls += uNum;
}

void Record::add_data(const uint32_t uNum) {
   uData += uNum;
}

void Record::add_total(const double dTime) {
   dTotal += dTime;
}

void Record::set_ident(const pair<HWND, uint32_t> pNew) {
   pIdent = pNew;
}

void Record::set_calls(const uint32_t uNum) {
   uCalls = uNum;
}

void Record::set_total(const double dNew) {
   dTotal = dNew;
}

void Record::set_percent(const double dNew) {
   dPercent = dNew;
}

void Record::set_average(const double dNew) {
   dAvgPc = dNew;
}

void Record::set_min(const double dNew) {
   dMinPc = dNew;
}

void Record::set_max(const double dNew) {
   dMaxPc = dNew;
}

void Record::reset() {
   dMaxPc = -1;
   dMinPc = -1;
   uData = 0;
}

// Sample class ////////////////////////////////////////////////////////////////

Record Sample::rSamples[MAX_PROFILER_SAMPLES];
int Sample::iLastSample = -1;
int Sample::iOpenSamples = 0;
uint64_t Sample::dBegin = 0;
uint64_t Sample::dEnd = 0;
Profiler *Sample::pProfiler = 0;

void Sample::init(HWND hWnd, uint32_t uMsg) {
   pair<HWND, uint32_t> pIdent;
   pIdent.first = hWnd;
   pIdent.second = uMsg;
   if (pProfiler->is_running() == true) {
      int iCount = 0;
      int iStoreIndex = -1;
      bool bDone = false;
      for( ; iCount < MAX_PROFILER_SAMPLES; ++iCount) {
         if (!rSamples[iCount].is_valid()) {
            if (iStoreIndex < 0) {
               iStoreIndex = iCount;
            }
         } else {
            pair<HWND, uint32_t> pCurrent = rSamples[iCount].get_ident();
            if (pCurrent.first == pIdent.first &&
                pCurrent.second == pIdent.second) {
               // If a second one comes along, we can't profile it because one's
               // already running - have to drop it. :/
               if (rSamples[iCount].is_closed()) {
                  iSampleIndex = iCount;
                  iParentIndex = iLastSample;
                  iLastSample = iCount;
                  ++iOpenSamples;
                  rSamples[iCount].open();
                  rSamples[iCount].add_call();
                  rSamples[iCount].start();
                  dBegin = rSamples[iCount].get_start();
                  bDone = true;
                  break;
               }
            }
         }
      }
      if (!bDone) {
         assert(iStoreIndex >= 0 && "Profiler has run out of sample slots!");
         rSamples[iStoreIndex].make_valid();
         rSamples[iStoreIndex].set_ident(pIdent);
         iSampleIndex = iStoreIndex;
         iParentIndex = iLastSample;
         iLastSample = iStoreIndex;
         iOpenSamples++;
         rSamples[iStoreIndex].open();
         rSamples[iStoreIndex].set_calls(1);
         rSamples[iStoreIndex].set_total(0.0f);
         rSamples[iStoreIndex].start();
         if (iParentIndex < 0) {
            dBegin = rSamples[iStoreIndex].get_start();
         }
      }
   }
}

void Sample::shutdown() {
   if (pProfiler->is_running() == true) {
      uint64_t dEndTime = pProfiler->get_time();
      rSamples[iSampleIndex].close();
      uint64_t dTime = dEndTime - rSamples[iSampleIndex].get_start();
      dEnd = dEndTime;
      rSamples[iSampleIndex].add_total(dTime);
      iLastSample = iParentIndex;
      --iOpenSamples;
      // Merge sample into record. Recalculate averages.
      double dSample = rSamples[iSampleIndex].get_total();
      double dAverage = rSamples[iSampleIndex].get_average();
      uint32_t uData = rSamples[iSampleIndex].get_data();
      double dTemp = (dAverage * (double)uData) + dSample;
      rSamples[iSampleIndex].add_data();
      rSamples[iSampleIndex].set_average(dTemp / ++uData);
      if (rSamples[iSampleIndex].get_min() == -1 ||
          dSample < rSamples[iSampleIndex].get_min()) {
         rSamples[iSampleIndex].set_min(dSample);
      }
      if (rSamples[iSampleIndex].get_max() == -1 ||
          dSample > rSamples[iSampleIndex].get_max()) {
         rSamples[iSampleIndex].set_max(dSample);
      }
      rSamples[iSampleIndex].set_calls(0);
      rSamples[iSampleIndex].set_total(0);
   }
   pProfiler->record_resource_usage();
}

void Sample::reset(const pair<HWND, uint32_t> pIdent) {
   for (int iCount = 0; iCount < MAX_PROFILER_SAMPLES; ++iCount) {
      if (rSamples[iCount].is_valid() &&
          rSamples[iCount].get_ident() == pIdent) {
         rSamples[iCount].reset();
         break;
      }
   }
}

void Sample::reset_all() {
   for (int iCount = 0; iCount < MAX_PROFILER_SAMPLES; ++iCount) {
      if(rSamples[iCount].is_valid()) {
         rSamples[iCount].reset();
         if (rSamples[iCount].is_open() == false) {
            rSamples[iCount].make_invalid();
         }
      }
   }
}

uint32_t Sample::get_total_samples() {
   uint32_t uCounter = 0;
   for (int iCount = 0; iCount < MAX_PROFILER_SAMPLES; ++iCount) {
      if(rSamples[iCount].is_valid()) {
         uCounter += rSamples[iCount].get_data();
      }
   }
   return uCounter;
}

// Profiler class //////////////////////////////////////////////////////////////

uint64_t Profiler::uTicksPerSecond;

Profiler::Profiler() : bRunning(true) {
   uTicksPerSecond = get_cpu_speed();
   assert(uTicksPerSecond != 0 && "Couldn't calculate clock speed!");
   dStart = get_time();
   srand(dStart);
   uGdiObjects = get_current_gdi_objects();
   uUserObjects = get_current_user_objects();
}

Profiler::~Profiler() {
}

const bool Profiler::is_running() const {
   return bRunning;
}

const uint32_t Profiler::get_cpu_speed() { // Return clock speed /Hz.
   uint32_t uRetVal = 0;
   LARGE_INTEGER lFreq;
   LARGE_INTEGER lTicks;
   LARGE_INTEGER lValue;
   LARGE_INTEGER lStart;
   if (QueryPerformanceFrequency(&lFreq)) {
      QueryPerformanceCounter(&lTicks);
      lValue.QuadPart = lTicks.QuadPart + lFreq.QuadPart;
      lStart.QuadPart = Profiler::get_cycle_count();
      do {
         QueryPerformanceCounter(&lTicks);
      } while (lTicks.QuadPart <= lValue.QuadPart);
      uRetVal = Profiler::get_cycle_count() - lStart.QuadPart;
   }
   return uRetVal;
}

const uint64_t Profiler::get_ticks_per_second() {
   return uTicksPerSecond;
}

inline volatile uint64_t Profiler::get_cycle_count() {
   // The notorious RDTSC instruction. This gives the current processor cycle.
   register uint64_t uTSC asm("eax");
   asm volatile (".byte 15, 49" : : : "eax", "edx");
   return uTSC;
}

const uint64_t Profiler::get_time() {
   return get_cycle_count();
}

const double Profiler::get_total_run_time() const {
   double dNumerator = 0;
   if (false == bRunning) {
      dNumerator = dEnd - dStart;
   }
   return (dNumerator / (double)uTicksPerSecond);
}

void Profiler::calculate_stats() {
   bRunning = false;
   dEnd = get_time();
}

void Profiler::output(const StatsWriter *sWriter) {
   time_t tClock;
   time(&tClock);
   tm *tTm = localtime(&tClock);
   char cTime[255] = "";
   sprintf(cTime, "%4d%02d%02d%02d%02d", tTm->tm_year + 1900, tTm->tm_mon,
           tTm->tm_mday, tTm->tm_hour, tTm->tm_min);
   string sTime = cTime;
   ofstream oFile(sWriter->create_filename(sTime).c_str(),
                  ios::out | ios::trunc);
   if (oFile.good() == true) {
      char cPath[256] = "";
      GetModuleFileName(0, cPath, 256);
      oFile << sWriter->header(cPath, tTm);
      oFile << sWriter->graph("CPU Usage", rCpu, "% CPU", "redbar.png");
      // RAM = "working set", VMM = "private bytes"
      // These generated graphs only hold the last 100 items, to minimize
      // resource consumption.
      oFile << sWriter->graph("Physical Memory Usage", rRam, " bytes",
                              "greenbar.png");
      oFile << sWriter->graph("Virtual Memory Usage", rVmm, " bytes",
                              "dkgreenbar.png");
      oFile << sWriter->graph("GDI Objects", rGdi, " objects", "bluebar.png");
      oFile << sWriter->graph("User Objects", rUser, " objects",
                              "yellowbar.png");
      oFile << sWriter->sample_data();
      oFile << sWriter->other_nums();
      oFile << sWriter->footer();
      oFile.close();
   }
}

const uint64_t Profiler::filetime_to_uint64_t(const FILETIME fTime) {
   uint64_t uTemp = fTime.dwHighDateTime;
   uTemp <<= 32;
   uTemp += fTime.dwLowDateTime;
   return uTemp;
}

const double Profiler::get_current_cpu_usage() {
   FILETIME fDummy = {0, 0};
   FILETIME fKernel = {0, 0};
   FILETIME fUser = {0, 0};
   FILETIME fSysKernel = {0, 0};
   FILETIME fSysUser = {0, 0};
   GetProcessTimes(GetCurrentProcess(), &fDummy, &fDummy, &fKernel, &fUser);
   GetSystemTimes(&fDummy, &fSysKernel, &fSysUser);
   uint64_t uCpuKernel = filetime_to_uint64_t(fSysKernel) - uLastCpuKernel;
   uint64_t uCpuUser = filetime_to_uint64_t(fSysUser) - uLastCpuUser;
   uint64_t uKernel = filetime_to_uint64_t(fKernel) - uLastKernel;
   uint64_t uUser = filetime_to_uint64_t(fUser) - uLastUser;
   uint64_t uSystem = uCpuKernel + uCpuUser;
   uint64_t uProcess = uKernel + uUser;
   uLastCpuUser = uCpuUser;
   uLastCpuKernel = uCpuKernel;
   uLastUser = uUser;
   uLastKernel = uKernel;
   return 100.0f * (long double)uProcess / (long double)uSystem;
}

const uint32_t Profiler::get_current_ram_usage() {
   PROCESS_MEMORY_COUNTERS pCount = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
   GetProcessMemoryInfo(GetCurrentProcess(), &pCount, sizeof(pCount));
   // There are lots of other interesting items in PROCESS_MEMORY_COUNTERS.
   // A future extension could explore some.
   return pCount.WorkingSetSize;
}

const uint32_t Profiler::get_current_vmm_usage() {
   PROCESS_MEMORY_COUNTERS pCount = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
   GetProcessMemoryInfo(GetCurrentProcess(), &pCount, sizeof(pCount));
   return pCount.PagefileUsage;
}

const uint16_t Profiler::get_current_gdi_objects() {
   return GetGuiResources(GetCurrentProcess(), GR_GDIOBJECTS);
}

const uint16_t Profiler::get_current_user_objects() {
   return GetGuiResources(GetCurrentProcess(), GR_USEROBJECTS);
}

const uint16_t Profiler::get_stored_gdi_objects() {
   return uGdiObjects;
}

const uint16_t Profiler::get_stored_user_objects() {
   return uUserObjects;
}

void Profiler::record_resource_usage() {
   rCpu << get_current_cpu_usage();
   rRam << get_current_ram_usage();
   rVmm << get_current_vmm_usage();
   rGdi << get_current_gdi_objects();
   rUser << get_current_user_objects();
}
